#ifndef lint
static const char rcsid[] =
	"$Id: poppwd.c,v 1.1.1.1 2001/01/12 06:06:44 andrew Exp $";
#endif

#include <sys/time.h>

#include <assert.h>
#include <errno.h>
#include <signal.h>
#include <stdarg.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <sysexits.h>
#include <syslog.h>
#include <unistd.h>

#include "config.h"
#include "poppwd.h"
#include "protocol.h"

static void usage(void);
static void version(void);
static void handle_sig_alarm(int unused);

/* flag set by alarm handler (used for timeouts) */
static volatile sig_atomic_t g_alarm_expired = 0;

int main(int argc, char *argv[]) {
	int c;

	/* set up logging */
	openlog("poppwd", LOG_PID, SYSLOG_FACILITY);

	/* handle command line arguments */
	while ((c = getopt(argc, argv, "v")) != -1) {
		switch (c) {
		case 'v':
			/* print version string and exit */
			version();
			break;
		case '?':
		default:
			usage();
			break;
		}
	}
	argc -= optind;
	argv += optind;

	/* make sure messages to client aren't buffered more than needed */
	if (setlinebuf(stdout) != 0) {
		basic_err(EX_UNAVAILABLE, MSG_ERR_MISC, "setlinebuf failed: %m");
	}

	/* do the actual eudora protocol */
	do_protocol();

	return(0);
}

/* print usage info and exit */
static void usage(void) {

	fprintf(stderr, "usage: poppwd [-v]\n");
	exit(EX_USAGE);
}

/* print version info and exit */
static void version(void) {

	printf("poppwd %s\n%s\n", VER_NUM, VER_COPYRIGHT);
	exit(EX_OK);
}

/* log an error to syslog, send an error message to the client and exit */
void basic_err(const int ev, const enum message_name message, const char *const format, ...) {
	va_list ap;

	va_start(ap, format);
	vsyslog(LOG_ERR, format, ap);
	va_end(ap);

	(void)to_client(TIME_OUT, "%s", messages[message]);

	exit(ev);
}

void basic_to_client(const enum message_name name) {

	if (to_client(TIME_OUT, "%s", messages[name]) < (signed int)(strlen(messages[name]) + 2)) {       
		/* client has probably died */
		basic_err(EX_IOERR, MSG_ERR_IO, "printf of %s failed: %m", messages[name]);
	}
}

/* send a message to the client. terminate all lines with \r\n.
 * returns number of characters printed
 *
 * TODO: implement timeout
 */
int to_client(unsigned int timeout, const char *const format, ...) {
	int rv;
	va_list ap;

	/* print what they said */
	va_start(ap, format);
	rv = vprintf(format, ap);
	va_end(ap);

	/* if that worked then terminate the line */
	if (rv > 0) {
		rv += printf("\r\n");
	}

	return(rv);
}

/* implement a read with a timeout.
 * 
 * timeout is the number of seconds to wait for input
 * buffer is a pointer to some memory where the read data can be stored
 * size is the size of the memory pointed to by buffer (and the maximum
 *	amount of data that will be read + 1)
 *
 * The timeout is implemented using alarm(3). This makes us susceptible to the
 * race condition as descriped in APUE 10.10. There is no simple way around
 * this and so weighing up the likely hood of this occuring and the severity of
 * it occuring I've decided to ignore it.
 *
 * This function stamps all over anyone elses SIGALRM setup. This isn't a
 * problem here. There should be a better way to timeout system calls.
 *
 * returns 0 on error, -1 on timeout, 1 on success
 */
int from_client(unsigned int timeout, unsigned int size, char *buffer) {
	struct sigaction sa;
	int read_rv;

	/* make sure SIGALRM interrupts read */
	sa.sa_handler = handle_sig_alarm;
	sigemptyset(&sa.sa_mask);
	sa.sa_flags = 0;
	if (sigaction(SIGALRM, &sa, NULL) != 0) {
		basic_err(EX_OSERR, MSG_ERR_MISC, "sigaction(SIGALRM) failed: %m");
	}

	/* schedule the alarm to be delivered */
	alarm(timeout);

	/* read data from client */
	while (size > 1) {
		if ((read_rv = read(STDIN_FILENO, buffer, (size_t)size - 1)) == -1) {
			/* error from read */
			if (errno == EINTR && g_alarm_expired) {
				/* timeout */
				basic_err(EX_OK, MSG_ERR_TIME, "read timeout");
			} else {
				basic_err(EX_OK, MSG_ERR_IO, "read failed: %m");
			}
		} else if (read_rv == 0) {
			/* EOF */
			basic_err(EX_OK, MSG_ERR_IO, "EOF unexpected");
		} else {
			/* get ready for next read */
			size -= read_rv; /* our buffer can now fit less */
			buffer += read_rv;	/* fill the buffer after the data we already have */
			if (*(buffer - 1) == '\n') {
				/* we have an entire line so we are done */
				/* disable any alarm that may still be running */
				alarm(0);
				/* null terminate the string */
				*buffer = '\0';
				/* indicate success */
				return(1);
			}
		}
	}

	/* we ran out of buffer space */
	basic_err(EX_PROTOCOL, MSG_ERR_TOOLONG, "out of buffer in from_client");

	/* NOT REACHED */
	assert(0);
	return(0);
}

static void handle_sig_alarm(int unused) {

	/* this space was intentionally left blank but then I put something here */
	g_alarm_expired = 1;	/* alarm_expired has file scope */

	/* keep gcc quiet */
	(void)unused;

	return;
}
